#include <termios.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <time.h>
#include <assert.h>
#include "nmea.h"
#include "com.h"
#include "macros.h"

#define XON  0x11
#define XOFF 0x13
#define BUFFER_CHUNK 128

#define FAKE_PORT "/dev/null"
//#define OS_FLOWCONTROL 

/* remember the port settings (_save) for later
 * restauration (_restore)
 */
static struct {
    const char *portname;
    FILE *port;
    struct termios settings;
} _saved;

/* save and restore the settings of the (opened) port.
 * _save() will register _restore() for usage atexit()
 */
static void _save(const char*, FILE*);
static void _restore();

FILE*
setup_port(const char* portname)
{
    struct termios settings;
    FILE *port = fopen(portname, "a+");

    if (strcmp(portname, FAKE_PORT) == 0) return port;
    if (!port) SYSDIE("open %s", portname);
    _save(portname, port);

    if (FAILED(tcgetattr(fileno(port), &settings)))
	SYSDIE("tcgetattr on %s", portname);

    cfsetispeed(&settings, B57600);
    cfsetospeed(&settings, B57600);
    cfmakeraw(&settings);
    settings.c_cflag |= CLOCAL | CS8;
#ifdef XONXOFF
    settings.c_iflag |= IXON;
#endif

    if (FAILED(tcsetattr(fileno(port), TCSANOW, &settings))) 
	SYSDIE("tcsetattr on %s", portname);

    return port;
}

void 
_send(FILE *port, const char* cmd)
{
    char *buffer;
    char *next;

    if (cmd[0] == '$') {
	buffer = alloca(strlen(cmd) + sizeof(",*ZZ\r\n"));
	next = buffer + sprintf(buffer, "%s,", cmd);
    }
    else {
	buffer = alloca(sizeof(PREFIX ",*ZZ\r\n") + strlen(cmd));
	next = buffer + sprintf(buffer, PREFIX "%s,", cmd);
    }

    next += sprintf(next, "*%02X\r\n", nmea_checksum(buffer));
    DEBUG DPRINTF("sending: %s\n", buffer);

    if (fputs(buffer, port) == EOF)
	SYSDIE("writing %s to device", buffer);
    tcdrain(fileno(port));
}

const char*
_recv(FILE *port)
{
    static struct {
	char *data;
	size_t size;
    } buffer;


    char c, *p;
    struct timespec t0, t1;
    int xoff_seen;

    /* avoid grown buffer by resetting it to some
     * small size */
    buffer.data = realloc(buffer.data, BUFFER_CHUNK);
    buffer.size = BUFFER_CHUNK;

    p = buffer.data;
    *p = '\0';
    xoff_seen = 0;


    while ((c = fgetc(port)) != EOF) {
	DEBUG { if (isprint(c)) DPRINTF("%c", c);
	      else if (c == XON) DPRINTF("<XON>");
	      else if (c == XOFF) DPRINTF("<XOFF>");
	      else DPRINTF("<%02x>", c);
	}

	//assert(c != '\0');

	if (isprint(c) || isspace(c)) { 
	    assert(buffer.size);
	    if (p == buffer.data + buffer.size - 2) {
		int offset = p - buffer.data;
		buffer.data = realloc(buffer.data, buffer.size += BUFFER_CHUNK);
		p = buffer.data + offset;
	    }
	    *p++ = c;
	}
	/* else WARN("skipping unprintable character <%02X>:", c); */


	/* process the character to detect if we have a complete
	 * sentence - we put '\0' into it to leave the loop */

	switch(c) {

#ifdef OS_FLOWCONTROL
	    /* When using software flow control of the OS a single record
	     * (sentence) ends with a new line
	     */
	    case '\n':	c = EOF; break;
#else
	    /* When using our own software flow control: as soon
	     * as we encounter an '\n' without an XOFF, somebody sent us
	     * a sentence and we're done. I think, this is a VIOLATION
	     * of the protocol and should never happen!
	     */
	    case '\n':	if (!xoff_seen) { 
			    WARN("protocol violation: <%s>\n", buffer.data);
			    c = EOF;
			}
			break; 
	    /* XOFF we get as soon as the internal processing of our
	     * data stats, it should be followed by the result
	     * sentenc(es) and an XON at the very end */
	    case XOFF:	xoff_seen = 1;
			DEBUG {
			    clock_gettime(CLOCK_REALTIME, &t0);
			    DPRINTF("- got XOFF\n");
			}
			break;
	    /* XON we get after successful command completion AND
	     * after successfull transmission of the data - if we get
	     * this XON we're done and my lave the loop, if, and only
	     * if we saw an XOFF before */
	    case XON:	DEBUG {
			    long int delta;
			    clock_gettime(CLOCK_REALTIME, &t1);
			    delta = (t1.tv_sec * 1000000000 + t1.tv_nsec) 
				- (t0.tv_sec * 1000000000 + t0.tv_nsec);
			    DPRINTF("- got XON after %.3fms (%d chars)\n",
				    (double)delta / 1000000,
				    xoff_seen - 1);
			}
			if (xoff_seen) c = EOF;
			break;
	    default:	if (xoff_seen) xoff_seen++;
#endif
	}
	if (c == EOF) break;
    }

    /* now terminate the buffer, advance to the first
     * ',', verify the checksum, cut the checksum and then
     * return the pointer to the start of the information.
     * NOTE: when we receive answers of several sentences
     * we need to do something more here!
     */
    *p = '\0';

    /* NMEA? */
    if (*buffer.data == '$') {
	if (!nmea_verify(buffer.data)) DIE("Checksum error: `%s'", buffer.data);
	if ((p = strchr(buffer.data, '*'))) *p = '\0';	/* cut the checksum */
	else DIE("Invalid NMEA sentence: `%s'", buffer.data);
	if ((p = strchr(buffer.data, ','))) return ++p;	/* cut the head (echo) */
	else DIE("Invalid NMEA sentence: `%s'", buffer.data);
    }

    return buffer.data;
}

const char*
__unused_recv(FILE *port)
{
    static char *buffer;
    static size_t bufsz;
    char chksum[3];
    ssize_t got;
    char *p;

    if ((got = getline(&buffer, &bufsz, port)) < 0)
	SYSDIE("getline");

    snprintf(chksum, sizeof(chksum), "%02X", nmea_checksum(buffer));

    DEBUG DPRINTF("<- <%s>\n", buffer);

    if ((p = strchr(buffer, '\r'))) *p = '\0';
    if ((p = strchr(buffer, '*'))) *p++ = '\0';

    if (strcmp(p, chksum) != 0) 
	DIE("checksum error");

    p = buffer;
    while (p && *p++ != ',')
	;

    return p;
}

const char*
talk(FILE* port, const char* cmd)
{
    _send(port, cmd);
    return _recv(port);
}

void _save(const char *portname, FILE* port)
{
    if (FAILED(tcgetattr(fileno(port), &_saved.settings))) 
	SYSDIE("tcgetattr on %s", portname)

    _saved.portname = portname;
    _saved.port = port;
    atexit(_restore);
}

static void 
_restore()
{
    if (!_saved.port) return;
    DEBUG DPRINTF("restoring port settings on %s\n", _saved.portname);

    if (FAILED(tcsetattr(fileno(_saved.port), TCSANOW, &_saved.settings)))
        SYSWARN("tcsetattr on %s", _saved.portname);

    DEBUG fputs("done\n", stderr);
}
