#include <assert.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include "competino.h"
#include "com.h"

#define CHAR 1
#define INT 2

#define ITEM(name, description, type, size...) \
    { name, #name, description, type, size }

static description_t eeprom[] = {
    #include "competino.c.eeprom"
    ITEM(_LAST_,	    "",				     0,          0, "", ""),
};

const description_t *_description_byname(const char*);
const description_t *_description_byaddr(enum options);

option_t _new_option(const description_t*, const char *value);
void destroy_option(option_t);

option_t _read_option(FILE *port, option_t);
void _write_option(FILE *port, option_t);

void _read(FILE *port, int addr, char *buffer, int need);
void _write(FILE *port, int addr, char *buffer, int size);

option_t* 
read_eeprom(FILE *port)
{
    option_t *rc = malloc(sizeof(eeprom) / sizeof(eeprom[0]) * sizeof(option_t));
    option_t *p;
    const description_t *d = eeprom;

    assert(rc);

    for (p = rc; d->address != _LAST_; ++d, ++p) {
	p->description = d;
	*p = _read_option(port, *p);
    }
    p->description = NULL;
    return rc;
}

option_t
read_option_byaddr(FILE *port, enum options address)
{
    option_t option = _new_option(_description_byaddr(address), NULL);
    return _read_option(port, option);
}

option_t
read_option_byname(FILE *port, const char *name)
{
    option_t option = _new_option(_description_byname(name), NULL);
    return _read_option(port, option);
}

void
write_option_byname(FILE *port, const char *name, const char *value)
{
    option_t option = _new_option(_description_byname(name), value);
    _write_option(port, option);
    destroy_option(option);
}

option_t
_new_option(const description_t *description, const char *value)
{

    option_t option;

    char *e = NULL;
    const description_t *error = NULL;

    option.description = description;

    if (value == NULL) {
	switch (option.description->type) {
	    case CSTRING:   option.value.cstring = NULL; break;
	    default:        option.value.number = 0; break;
	}
	return option;
    }

    /* conversion string to option value */
    errno = 0;
    switch (option.description->type) {
	case CSTRING: option.value.cstring = malloc(option.description->size);
		      assert(option.value.cstring);
		      snprintf(option.value.cstring, option.description->size, "%-15s", value);
		      DEBUG DPRINTF("cstring: `%s'\n", option.value.cstring);
		      break;
	case SIGNED:  option.value.number = strtol(value, &e, 10);
		      DEBUG DPRINTF("signed: %d\n", option.value.number);
		      break;
	case UNSIGNED:option.value.number = strtoul(value, &e, 10);
		      DEBUG DPRINTF("unsigned: %d\n", option.value.number);
		      break;
    }

    /* check the option value */
    if (e && !*e) e = NULL;
    switch (option.description->type) {
	case CSTRING: if (strlen(option.value.cstring) >= option.description->size)
			error = option.description;
		      break;
	case SIGNED:  if (e 
			    || errno 
			    || abs(option.value.number) >= (1 << (option.description->size * 8 - 1)))
			error = option.description;
		      break;
	case UNSIGNED:if (e 
			    || errno 
			    || option.value.number < 0 
			    || option.value.number >= (1 << option.description->size * 8))
			error = option.description;
		      break;
    }
    if (error) {
	char *text = NULL;
	if (error->remarks) {
	    text = malloc(sizeof(" ()") + strlen(error->remarks));
	    sprintf(text, " (%s)", error->remarks);
	}
	DIE("option %s: bad value `%s' [%s | %s%s]", error->name, value, 
	    error->text,
	    error->range ? error->range : "",
	    text ? text : "");
	free(text);
    }

    return option;
}

void
destroy_option(option_t o)
{
    if (o.description->type == CSTRING) free(o.value.cstring);
}

void
_write_option(FILE *port, option_t option)
{

    char *buffer = malloc(option.description->size);
    char *d = buffer;
    int i;
    assert(buffer);

    switch(option.description->type) {

	case CSTRING:	d = buffer + snprintf(buffer, option.description->size, "%-*s", 
			    (int)option.description->size - 1,
			    option.value.cstring);
			break;
	case SIGNED:
	case UNSIGNED:  for (i = option.description->size, d = buffer; i; --i) 
			    *d++ = 0xFF & (option.value.number >> (i - 1) * 8);
			*d = '\0';
			break;
    }

    /* fprintf(stderr, "%d|%s|\n", d-buffer + 1, buffer); */
    _write(port, option.description->address, buffer, d - buffer + 1);

    free(buffer);
}

const description_t*
_description_byaddr(enum options address)
{
    const description_t *d;
    for (d = eeprom; (d->address != _LAST_) && (d->address != address); ++d)
	;

    if (d->address == _LAST_)
	DIE("unknown address %d\n", address);

    return d;
}

const description_t*
_description_byname(const char *name)
{
    const description_t *d;
    for (d = eeprom; (d->address != _LAST_) && (strcmp(d->name, name)); ++d)
	;
    if (d->address == _LAST_)
	DIE("unknown option %s\n", name);

    return d;
}


void
_write(FILE *port, int addr, char *buffer, int size)
{
    char cmd[] = "MEMW,XXXX,N,a1,a2,a3,a4,a5,a6,a7,a8";
    char *d = d;
    char *p = buffer;

    assert(size);

    DEBUG DPRINTF("sending %d chars\n", size);

    for (;;) {
	if ((p - buffer) % 8 == 0) 
	    d = cmd + sprintf(cmd, "MEMW,%04X,%d", 
		addr + (int)(p - buffer), 
		size > 8 ? 8 : size);
	
	d += sprintf(d, ",%02hhX", *p++);
	--size;

	if (size == 0) {
	    int missing;
	    for (missing = (p - buffer) % 8; missing; --missing)
		d+= sprintf(d, ",");
	    (void) talk(port, cmd);
	    break;
	}

	if (((p - buffer) % 8) == 0) {
	    (void) talk(port, cmd);
	    continue;
	}

    }


}

void 
_read(FILE *port, int addr, char *buffer, int need)
{
    static struct {
        const int size;
        int addr;
        char *buffer;
    } cache = { 8, -1, NULL };

    struct {
        int addr;
        char *buffer;
        int size;
    } chunk = { addr, buffer, need };

    DEBUG DPRINTF("need %2d bytes from 0x%04X\n", need, addr);

    while (chunk.size) {
        const char *p;
        char *d;

        DEBUG DPRINTF(": chunk: 0x%04X + %d bytes\n", chunk.addr, chunk.size);

        /* If the cache is empty or if the byte range we need is totally
         * outside of the cache, we need to fill it. 
         */
        if (!cache.buffer
            || (chunk.addr + chunk.size <= cache.addr) 
            || (chunk.addr >= cache.addr + cache.size)) {

            char cmd[sizeof("MEMR,XXXX")];
            char *e;

            cache.buffer = malloc(8);
            assert(cache.buffer);

            DEBUG DPRINTF(": cache MISS 0x%04X\n", chunk.addr);

            sprintf(cmd, "MEMR,%04X", chunk.addr);
            p = talk(port, cmd);

            /* now we've something like 0020,11,22,33,44,55,66,77,88<00>
             * and need to transform it into single characters
             */
            assert(p);
            assert(strlen(p) == sizeof("0000,11,22,33,44,55,66,77,88") - 1);

            /* parse the address - should fit the request and then
             * advance until the first ',' */
            assert(strtoul(p, NULL, 0x10) == (unsigned) chunk.addr);
            while (*p && (*p++ != ',')) 
                ;
            assert(*p && *p != '\0');

            /* fill our cache first */
            cache.addr = chunk.addr;
            for (d = cache.buffer; 
                    *p && d < cache.buffer + cache.size;
                    p = *e ? e + 1 : e) {
                *d++ = strtoul(p, &e, 0x10);
                assert(*e == ',' || *e == '\0');
            }

            DEBUG {
                const char *p;
                DPRINTF(": cache 0x%04X: ", cache.addr);
                for (p = cache.buffer; p < cache.buffer + cache.size; ++p) DPRINTF("%02hhX ", *p);
                for (p = cache.buffer; p < cache.buffer + cache.size; ++p) DPRINTF("%c", isgraph(*p) ? *p : '.');
                DPRINTF("\n");
            }

        } else DEBUG DPRINTF(": cache HIT 0x%04X\n", chunk.addr);

        /* and now copy the needed bytes to the buffer */
        if (cache.addr < chunk.addr) {
	    DEBUG DPRINTF(":: %03x < %03x | %03x\n", cache.addr, chunk.addr, addr);
            p = cache.buffer + (chunk.addr - cache.addr);
            d = buffer + (chunk.addr - addr);
        } 
	else if (cache.addr > chunk.addr) {
	    DEBUG DPRINTF(":: %03x > %03x | %03x\n", cache.addr, chunk.addr, addr);
            p = cache.buffer;
            d = buffer + (cache.addr - addr);
        } 
	else {
	    DEBUG DPRINTF(":: %03x ? %03x | %03x\n", cache.addr, chunk.addr, addr);
            p = cache.buffer;
            d = buffer + (chunk.addr - addr);
        }

        while ((p < cache.buffer + cache.size) 
            && d < (buffer + (chunk.addr - addr) + chunk.size)) {

            DEBUG DPRINTF(": buffer[%ld] = cache[%ld] %c\n", d - buffer, p - cache.buffer,
                isgraph(*p) ? *p : '.');
            *d++ = *p++;
        }

	DEBUG {
	    DPRINTF(":       addr: %04X + %d\n", addr, need);
	    DPRINTF(": chunk.addr: %04X + %d\n", chunk.addr, chunk.size);
	    DPRINTF(": cache.addr: %04X\n", cache.addr);
	    DPRINTF(": distance  : %ld\n", d - buffer);
	}


        /* now several cases we've to separate
         * 1) the cache was at the beginning of the buffer
         * 2) the cache was at the end of the buffer
         * 3) the cache was just somewhere in the middle of the buffer
         *    (should not happen currently)
         */

	DEBUG DPRINTF(": - adjusting -\n");
        if (cache.addr <= chunk.addr) {
            chunk.addr = addr + d - buffer;
            chunk.size = buffer + need - d;
            DEBUG {
		DPRINTF(": cache.addr <= chunk.addr\n");
		DPRINTF(": chunk.addr: %04X + %d\n", chunk.addr, chunk.size);
	    }
        } 
        else if (cache.addr > chunk.addr && ((cache.addr + cache.size) > (addr + need))) {
            chunk.addr = addr;
            chunk.size = chunk.addr - addr;
            DEBUG {
		DPRINTF(": cache.addr > chunk.addr ...\n");
		DPRINTF(": chunk.addr: %04X + %d\n", chunk.addr, chunk.size);
	    }
        } 
        else {
            DEBUG DPRINTF(": data <cache> data \n");
            _read(port, addr, buffer, cache.addr - addr);
            _read(port, addr + (d - buffer), d, buffer + need - d);
	    return;
        }

    }

    return;

}

option_t
_read_option(FILE *port, option_t option)
{
    const description_t *d = option.description;	/* just for convenience */
    DEBUG DPRINTF("requesting %2ld bytes for %s from address %04d/0x%04X\n", 
	d->size, d->name, d->address, d->address);

    /* only for strings we need to malloc the space, the plain numbers
     * are stored directly. The vario uses little endian for the 2byte
     * values, our intel architecture uses little endian too, so we do
     * not need to adjust anything. */
    if (d->type == CSTRING) {
        option.value.cstring = malloc(d->size);
        assert(option.value.cstring);
    }

    _read(port, d->address, 
                d->type == CSTRING ? (char*)option.value.cstring : (char*)&option.value.number,
                d->size);

    /* byte swap, the eeprom is BIG endian (in contrast to the
     * statement in the documentation) */
    if (d->size == INT)
        option.value.number = (0x100 * (option.value.number & 0xFF)) 
                             + (option.value.number >> 8);
    return option;
}

const char*
option_value(option_t o)
{
    static char *buffer;

    switch(o.description->type) {
	case CSTRING:	buffer = realloc(buffer, strlen(o.value.cstring + sizeof("")));
			sprintf(buffer, "\"%-*s\"", (int)o.description->size-1, o.value.cstring);
			break;
	case SIGNED:	buffer = realloc(buffer, 1024);
			sprintf(buffer, "%3d", o.value.number); 
			break;
	case UNSIGNED:  buffer = realloc(buffer, 1024);
			sprintf(buffer, "%3u", o.value.number); 
			break;
	default:        DIE("unknown type %d\n", o.description->type);
    }

    return buffer;
}

void
print_option(FILE* p, option_t o)
{

    fprintf(p, "%-20s: %5s [%s", o.description->name, option_value(o), o.description->text);
    if (o.description->range && *(o.description->range)) fprintf(p, "|%s", o.description->range);
    if (o.description->remarks && *(o.description->remarks)) fprintf(p, "|%s", o.description->remarks);
    fprintf(p, "]\n");
}
